@@ -5,15 +5,15 @@ Created on 2023-06-11
 """
 import json
 import os
-import yaml
-from dataclasses import asdict, dataclass, field
+from dataclasses import dataclass, field
 from json.decoder import JSONDecodeError
-from typing import Dict, List, Optional, Tuple, Union
-from urllib.parse import quote_plus
+from typing import Any, Dict, List, Optional, Tuple, Union
 
 import markdown2
+import yaml
 from dataclasses_json import dataclass_json
-from ngwidgets.yaml import Yaml
+from ngwidgets.yamlable import YamlAble
+from slugify import slugify
 
 from dcm.svg import SVG, SVGConfig, SVGNodeConfig
 
@@ -38,6 +38,11 @@ class CompetenceElement:
     description: Optional[str] = None
     color_code: Optional[str] = None
 
+    def __post_init__(self):
+        # Set the id to the the slug of the name if id is None
+        if self.id is None:
+            self.id = slugify(self.name)
+
     def as_html(self) -> str:
         """
         convert me to html
@@ -86,23 +91,28 @@ class CompetenceFacet(CompetenceElement):
     This class can include additional properties or methods specific to a competence facet.
     """
 
-    def __post_init__(self):
-        # Set the id to the name if id is None
-        if self.id is None:
-            self.id = self.name
+@dataclass_json
+@dataclass
+class CompetenceArea(CompetenceElement):
+    """
+    Represents a specific area within a competence aspect, containing various facets.
 
+    Attributes:
+        facets (List[CompetenceFacet]): A list of CompetenceFacet objects representing individual facets of this area.
+    """
+    facets: List[CompetenceFacet] = field(default_factory=list)
 
 @dataclass_json
 @dataclass
 class CompetenceAspect(CompetenceElement):
     """
-    Represents a broader category of competence, which includes various facets.
+    Represents a broader category of competence, which includes various areas.
 
     Attributes:
-        facets (List[CompetenceFacet]): A list of CompetenceFacet objects representing individual facets of this aspect.
+        areas (List[CompetenceArea]): A list of CompetenceArea objects representing individual areas of this aspect.
     """
 
-    facets: List[CompetenceFacet] = field(default_factory=list)
+    areas: List[CompetenceArea] = field(default_factory=list)
     credits: Optional[int] = None
 
 
@@ -114,40 +124,111 @@ class CompetenceLevel(CompetenceElement):
 
     Attributes:
         level (int): level number starting from 1 as the lowest and going up to as many level as defined for the CompetenceTree
+        icon(str): the name of a google mdi icon to be shown for this level
+        utf8_icon(str): utf-8 char string to be used as icon
     """
 
     level: int = 1
+    icon: Optional[str] = None
+    utf8_icon: Optional[str] = None
 
 
 @dataclass_json
 @dataclass
-class CompetenceTree(CompetenceElement):
+class CompetenceTree(CompetenceElement, YamlAble["CompetenceTree"]):
     """
     Represents the entire structure of competencies, including various aspects and levels.
 
     Attributes:
-        competence_aspects (Dict[str, CompetenceAspect]): A dictionary mapping aspect IDs to CompetenceAspect objects.
+        competence_aspects (List[CompetenceAspect]): A list of CompetenceAspect objects.
         competence_levels (List[CompetenceLevel]): A list of CompetenceLevel objects representing the different levels in the competence hierarchy.
         element_names (Dict[str, str]): A dictionary holding the names for tree, aspects, facets, and levels.  The key is the type ("tree", "aspect", "facet", "level").
     """
 
     lookup_url: Optional[str] = None
-    competence_aspects: Dict[str, CompetenceAspect] = field(default_factory=dict)
-    competence_levels: List[CompetenceLevel] = field(default_factory=list)
+    aspects: List[CompetenceAspect] = field(default_factory=list)
+    levels: List[CompetenceLevel] = field(default_factory=list)
     element_names: Dict[str, str] = field(default_factory=dict)
 
     def __post_init__(self):
-        # prepare yaml dumping
-        self.yaml=Yaml()
-        for aspect_id, aspect in self.competence_aspects.items():
-            if aspect.id is None:
-                aspect.id = aspect_id
+        """
+        initalize the path variables of my hierarchy
+        """
+        super().__post_init__()
+        self.path = self.id
+        # Loop through each competence aspect and set their paths and parent references
+        for aspect in self.aspects:
+            aspect.competence_tree = self
+            aspect.path = f"{self.id}/{aspect.id}"
+            for area in aspect.areas:
+                area.competence_tree = self
+                area.aspect = aspect
+                area.path = f"{self.id}/{aspect.id}/{area.id}"
+                for facet in area.facets:
+                    facet.competence_tree = self
+                    facet.area = area
+                    facet.path = f"{self.id}/{aspect.id}/{area.id}/{facet.id}"
 
     @classmethod
     def required_keys(cls) -> Tuple:
         keys = {"name", "id", "url", "description", "element_names"}
         return keys
 
+    def lookup_by_path(
+        self, path: str, lenient: bool = True
+    ) -> Optional[CompetenceElement]:
+        """
+        Look up and return a competence element (tree,aspect of facet)
+        based on the given path.
+
+        The path is expected to be in the format "tree_id/aspect_id/facet_id".
+        This method parses the path and retrieves the corresponding competence aspect or facet.
+
+        Args:
+            path (str): The path in the format "tree_id/aspect_id/facet_id".
+
+            lenient(bool): if not lenient raise Exceptions for invalid paths and ids
+        Returns:
+            Optional[CompetenceElement]: The competence aspect or facet corresponding to the given path.
+        """
+
+        def handle_error(msg):
+            if not lenient:
+                raise ValueError(msg)
+
+        parts = path.split("/")
+        if len(parts) < 1:
+            return None
+
+        tree_id = parts[0]
+        if tree_id != self.id:
+            handle_error(f"invalid tree_id for lookup {tree_id}")
+            return None
+        if len(parts) == 1:
+            return self
+        if len(parts) > 1:
+            aspect_id = parts[1]
+            # Retrieve the aspect
+            aspect = next((aspect for aspect in self.aspects if aspect.id==aspect_id), None)
+        if aspect:
+            if len(parts) == 2:
+                return aspect
+            if len(parts) > 2:
+                area_id = parts[2]
+                area = next((area for area in aspect.areas if area.id == area_id), None)
+                if area:
+                    if len(parts) == 3:
+                        return area
+                if len(parts) > 3:
+                    facet_id = parts[3]
+                    facet = next(
+                        (facet for facet in area.facets if facet.id == facet_id), None
+                    )
+                    if facet:
+                        return facet
+        handle_error(f"invalid path for lookup {path}")
+        return None
+
     def to_pretty_json(self):
         """
         Converts the CompetenceTree object to a pretty JSON string, handling null values.
@@ -171,13 +252,6 @@ class CompetenceTree(CompetenceElement):
         null_free_json_str = json.dumps(none_free_dict, indent=2)
         return null_free_json_str
 
-    def to_yaml(self):
-        """
-        Converts the CompetenceTree object to a YAML string.
-        """
-        yaml_str=self.yaml.to_yaml(self)
-        return yaml_str
-
     def add_legend(self, svg: SVG) -> None:
         """
         Add a legend to the SVG explaining the color codes for levels and aspects.
@@ -194,9 +268,7 @@ class CompetenceTree(CompetenceElement):
         padding = 5
 
         # Add the competence level legend
-        level_items = [
-            (level.color_code, level.name) for level in self.competence_levels
-        ]
+        level_items = [(level.color_code, level.name) for level in self.levels]
         svg.add_legend_column(
             level_items,
             self.element_names.get("level", "Level"),
@@ -211,15 +283,12 @@ class CompetenceTree(CompetenceElement):
             x_start
             + box_width
             + padding
-            + max(svg.get_text_width(level.name) for level in self.competence_levels)
+            + max(svg.get_text_width(level.name) for level in self.levels)
             + padding
         )
 
         # Add the competence aspect legend
-        aspect_items = [
-            (aspect.color_code, aspect.name)
-            for aspect in self.competence_aspects.values()
-        ]
+        aspect_items = [(aspect.color_code, aspect.name) for aspect in self.aspects]
         svg.add_legend_column(
             aspect_items,
             self.element_names.get("aspect", "Aspect"),
@@ -237,40 +306,101 @@ class Achievement:
     Class representing an individual's achievement level for a specific competence facet.
 
     Attributes:
-        facet_id (str): Identifier for the competence facet.
+        path (str): The path in the CompetenceTree, used to derive tree_id, aspect_id, and facet_id.
         level (int): The achieved level for this facet.
-        percent(float): how well was the achievement reached?
+        score (float): How well the achievement was reached.
+        score_unit (str): Unit of the score, default is "%".
         evidence (Optional[str]): Optional evidence supporting the achievement.
         date_assessed (Optional[str]): Optional date when the achievement was assessed (ISO-Format).
     """
 
-    facet_id: str
-    level: int
-    score: float
+    path: str
+    level: int = None
+    score: float = None
     score_unit: Optional[str] = "%"
     evidence: Optional[str] = None
-    date_assessed: Optional[str] = None
+    date_assessed_iso: Optional[str] = None
+
+    @property
+    def tree_id(self):
+        parts = self.path.split("/")
+        return parts[0] if parts else None
+
+    @property
+    def aspect_id(self):
+        parts = self.path.split("/")
+        return parts[1] if len(parts) > 1 else None
+
+    @property
+    def area_id(self):
+        parts = self.path.split("/")
+        return parts[2] if len(parts) > 2 else None
+
+    @property
+    def facet_id(self):
+        parts = self.path.split("/")
+        return parts[3] if len(parts) > 3 else None
 
 
-@dataclass
 @dataclass_json
-class Student:
+@dataclass
+class Learner:
     """
-    A student with achievements.
+    A learner with achievements.
     Attributes:
-        student_id (str): Identifier for the student.
-        achievements (Dict[str, List[Achievement]]): A dictionary where each key is a competence tree identifier
-                                                     and the value is a list of Achievement instances for that tree.
+        learner_id (str): Identifier for the learner.
+        achievements (Dict[str, List[Achievement]]):
+            A dictionary where each key is a competence element identifier
+            and the value is a list of Achievement instances for that tree.
     """
 
-    student_id: str
-    achievements: Dict[str, List[Achievement]]
+    learner_id: str
+    achievements: Optional[List[Achievement]] = field(default=None)
+
+    def __post_init__(self):
+        self.achievements_by_path = {}
+        if self.achievements:
+            for achievement in self.achievements:
+                self.achievements_by_path[achievement.path] = achievement
 
     @classmethod
     def required_keys(cls):
         keys = {"achievements"}
         return keys
 
+    @property
+    def main_id(self):
+        main_id = self.learner_id
+        return main_id
+
+    def add_achievement(self, new_achievement):
+        self.achievements.append(new_achievement)
+        self.achievements_by_path[new_achievement.path] = new_achievement
+
+    def get_competence_tree_ids(self) -> List[str]:
+        """
+        Get all unique competence tree IDs of my achievements.
+
+        Returns:
+            List[str]: A list of unique competence tree IDs.
+        """
+        # Assuming that the learner's achievements are stored in a list called self.achievements
+        # You can modify this part according to your actual data structure.
+
+        # Create a set to store unique competence tree IDs
+        unique_tree_ids = set()
+
+        # Iterate through the learner's achievements
+        for achievement in self.achievements:
+            # Assuming each achievement has a tree_id attribute
+            tree_id = achievement.tree_id
+
+            # Add the tree_id to the set
+            unique_tree_ids.add(tree_id)
+
+        # Convert the set to a list and return
+        return list(unique_tree_ids)
+
 
 class DynamicCompetenceMap:
     """
@@ -284,37 +414,10 @@ class DynamicCompetenceMap:
         self.competence_tree = competence_tree
         self.svg = None
 
-    def lookup(
-        self, aspect_id: str = None, facet_id: str = None
-    ) -> Optional[CompetenceElement]:
-        """
-        Look up an element of the competence tree by aspect ID and facet ID.
-        - Returns the entire competence tree if both aspect_id and facet_id are None.
-        - Returns a specific aspect if only aspect_id is provided.
-        - Returns a specific facet within an aspect if both aspect_id and facet_id are provided.
-
-        Args:
-            aspect_id (str, optional): The ID of the aspect to search within.
-            facet_id (str, optional): The ID of the facet to find within the specified aspect.
-
-        Returns:
-            Optional[CompetenceElement]: The found competence aspect, facet, or the entire competence tree.
-        """
-        ct = self.competence_tree
-        if aspect_id is None:
-            return ct
-        aspect = ct.competence_aspects.get(aspect_id)
-        if aspect is None:
-            for ct_aspect in ct.competence_aspects.values():
-                if aspect.id == aspect_id:
-                    aspect = ct_aspect
-        if aspect:
-            if facet_id is None:
-                return aspect
-            for facet in aspect.facets:
-                if facet.id == facet_id:
-                    return facet
-        return None
+    @property
+    def main_id(self):
+        main_id = self.competence_tree.id
+        return main_id
 
     @classmethod
     def examples_path(cls) -> str:
@@ -362,7 +465,9 @@ class DynamicCompetenceMap:
                                 else:
                                     example_dcm_defs[file_prefix] = definition_data
                         except Exception as ex:
-                            cls.handle_markup_issue(filename, definition_text, ex)
+                            cls.handle_markup_issue(
+                                filename, definition_text, ex, markup
+                            )
         return example_dcm_defs
 
     @classmethod
@@ -381,14 +486,16 @@ class DynamicCompetenceMap:
             ValueError: If an unsupported markup format is specified.
         """
         if markup == "json":
-            return json.loads(text)
+            data=json.loads(text)
+            return data
         elif markup == "yaml":
-            return yaml.safe_load(text)
+            data=yaml.safe_load(text)
+            return data
         else:
             raise ValueError(f"Unsupported markup format: {markup}")
 
     @classmethod
-    def handle_markup_issue(cls, name: str, definition_string: str, ex, markup):
+    def handle_markup_issue(cls, name: str, definition_string: str, ex, markup: str):
         if isinstance(ex, JSONDecodeError):
             lines = definition_string.splitlines()  # Split the string into lines
             err_line = lines[ex.lineno - 1]  # JSONDecodeError gives 1-based lineno
@@ -416,18 +523,20 @@ class DynamicCompetenceMap:
         for name, definition_string in cls.get_example_dcm_definitions(
             required_keys=content_class.required_keys(), markup=markup
         ).items():
-            dcm = cls.from_definition_string(
+            example = cls.from_definition_string(
                 name, definition_string, content_class, markup=markup
             )
-            examples[dcm.competence_tree.id] = dcm
+            # check the type of the example
+            example_id = example.main_id
+            examples[example_id] = example
         return examples
 
     @classmethod
     def from_definition_string(
         cls, name: str, definition_string: str, content_class, markup: str = "json"
-    ) -> "DynamicCompetenceMap":
+    ) -> Any:
         """
-        Load a DynamicCompetenceMap instance from a definition string (either JSON or YAML).
+        Load a DynamicCompetenceMap or Learner instance from a definition string (either JSON or YAML).
 
         Args:
             name (str): A name identifier for the data source.
@@ -444,160 +553,9 @@ class DynamicCompetenceMap:
         try:
             data = cls.parse_markup(definition_string, markup)
             content = content_class.from_dict(data)
-            return DynamicCompetenceMap(content)
+            if isinstance(content, CompetenceTree):
+                return DynamicCompetenceMap(content)
+            else:
+                return content
         except Exception as ex:
             cls.handle_markup_issue(name, definition_string, ex, markup)
-
-    def generate_svg(
-        self, filename: Optional[str] = None, config: Optional[SVGConfig] = None
-    ) -> str:
-        """
-        Generate the SVG markup and optionally save it to a file. If a filename is given, the method
-        will also save the SVG to that file. The SVG is generated based on internal state not shown here.
-
-        Args:
-            filename (str, optional): The path to the file where the SVG should be saved. Defaults to None.
-            config (SVGConfig, optional): The configuration for the SVG canvas and legend. Defaults to default values.
-
-        Returns:
-            str: The SVG markup.
-        """
-        if config is None:
-            config = SVGConfig()  # Use default configuration if none provided
-        svg_markup = self.generate_svg_markup(self.competence_tree, config)
-        if filename:
-            self.save_svg_to_file(svg_markup, filename)
-        return svg_markup
-
-    def generate_svg_markup(
-        self,
-        competence_tree: CompetenceTree = None,
-        config: SVGConfig = None,
-        with_java_script: bool = True,
-        lookup_url: str = "",
-    ) -> str:
-        """
-        Generate SVG markup based on the provided competence tree and configuration.
-
-        Args:
-            competence_tree (CompetenceTree): The competence tree structure containing the necessary data.
-            config (SVGConfig): The configuration for the SVG canvas and legend.
-            lookup_url(str): the lookup_url to use if there is none defined in the CompetenceTree
-
-        Returns:
-            str: The generated SVG markup.
-        """
-        if competence_tree is None:
-            competence_tree = self.competence_tree
-        lookup_url = (
-            competence_tree.lookup_url if competence_tree.lookup_url else lookup_url
-        )
-        competence_aspects = competence_tree.competence_aspects
-        # Instantiate the SVG class
-        svg = SVG(config)
-        self.svg = svg
-        # use default config incase config was None
-        config = svg.config
-
-        # Center of the donut
-        # Center of the donut chart should be in the middle of the main SVG area, excluding the legend
-        cx = svg.width // 2
-        cy = (config.total_height - config.legend_height) // 2  # Adjusted for legend
-
-        # Calculate the radius for the central circle (10% of the width)
-        tree_radius = cx / 9
-
-        # Add the central circle representing the CompetenceTree
-        circle_config = competence_tree.to_svg_node_config(
-            x=cx, y=cy, width=tree_radius
-        )
-        svg.add_circle(config=circle_config)
-
-        facet_radius = min(cx, cy) * 0.9  # Leave some margin
-        aspect_radius = facet_radius / 3  # Choose a suitable inner radius
-
-        # Total number of facets
-        total_facets = sum(len(aspect.facets) for aspect in competence_aspects.values())
-
-        # Starting angle for the first aspect
-        aspect_start_angle = 0
-
-        for aspect_code, aspect in competence_aspects.items():
-            num_facets_in_aspect = len(aspect.facets)
-
-            # Skip aspects with no facets
-            if num_facets_in_aspect == 0:
-                continue
-            aspect_angle = (num_facets_in_aspect / total_facets) * 360
-
-            aspect_url = (
-                aspect.url
-                if aspect.url
-                else f"{lookup_url}/description/{quote_plus(competence_tree.id)}/{quote_plus(aspect_code)}"
-                if lookup_url is not None
-                else None
-            )
-            show_as_popup = aspect.url is None
-
-            aspect_config = aspect.to_svg_node_config(
-                url=aspect_url,
-                show_as_popup=show_as_popup,
-                x=cx,
-                y=cy,
-                width=tree_radius,  # inner radius
-                height=aspect_radius,  # outer radius
-            )
-            # fix id
-            aspect_config.id = aspect_code
-            # Draw the aspect segment as a donut segment
-            svg.add_donut_segment(
-                config=aspect_config,
-                start_angle_deg=aspect_start_angle,
-                end_angle_deg=aspect_start_angle + aspect_angle,
-            )
-
-            facet_start_angle = (
-                aspect_start_angle  # Facets start where the aspect starts
-            )
-            angle_per_facet = (
-                aspect_angle / num_facets_in_aspect
-            )  # Equal angle for each facet
-
-            for facet in aspect.facets:
-                # Add the facet segment as a donut segment
-                facet_url = (
-                    facet.url
-                    if facet.url
-                    else f"{lookup_url}/description/{quote_plus(competence_tree.id)}/{quote_plus(aspect_code)}/{quote_plus(str(facet.id))}"
-                    if lookup_url is not None
-                    else None
-                )
-                show_as_popup = facet.url is None
-                facet_config = facet.to_svg_node_config(
-                    url=facet_url,
-                    show_as_popup=show_as_popup,
-                    x=cx,
-                    y=cy,
-                    width=aspect_radius,  # inner radius
-                    height=facet_radius,  # outer radius
-                )
-                svg.add_donut_segment(
-                    config=facet_config,
-                    start_angle_deg=facet_start_angle,
-                    end_angle_deg=facet_start_angle + angle_per_facet,
-                )
-                facet_start_angle += angle_per_facet
-
-            aspect_start_angle += aspect_angle
-
-        # optionally add legend
-        if config.legend_height > 0:
-            self.competence_tree.add_legend(svg)
-
-        # Return the SVG markup
-        return svg.get_svg_markup(with_java_script=with_java_script)
-
-    def save_svg_to_file(self, svg_markup: str, filename: str):
-        # Save the SVG content to a file
-        with open(filename, "w") as file:
-            file.write(svg_markup)
